// file         : functions_more_operations.cpp
// filetype     : jjs-bdd demo file
// authors      : Sascha Klueppelholz, Joern Ossowski
// date         : 02-11-2004
// comment      : demo of additional operations on functions

#include <iostream>

#include <sobdd/sobdd.h>

using namespace std;

int main(int argc, char *argv[]){
  cout << endl << "Starting jjs-bdd demo for bdd operations on functions. Take a look at the source file for more details." << endl << endl;

  // the projection function to a specific variable can be added
  // if the new variable is not already in the variable order, it will be appended automaticly
  SOBDDFunction p_x=SOBDDFunction::projection("x");
  SOBDDFunction p_y=SOBDDFunction::projection("y");
  SOBDDFunction p_z=SOBDDFunction::projection("z");

  // as mentioned in our first example, ITE can be applied
  SOBDDFunction ite=ITE(p_x,p_y,p_z);

  // cofactors can be builded
  cout << "... negative cofactor of a projection function is ZERO()?: " << (p_x.cofactor("x",0)==SOBDDFunction::ZERO()) << endl;
  cout << "... and the positive cofactor is ONE()?: " << (p_x.cofactor("x",1)==SOBDDFunction::ONE()) << endl;

  // quantification is supported by the jjs-bdd package
  cout << "... exists(x) of projection on x is ONE()?: " << (p_x.exists("x")==SOBDDFunction::ONE()) << endl;
  cout << "... while forall(x) is ZERO()?: " << (p_x.forall("x")==SOBDDFunction::ZERO()) << endl;

  // quantification with sets of variables is also supported
  VarSet varset;
  varset.insert("x");
  varset.insert("y");
  cout << "... exists(x,y) of AND(x,AND(y,z)) is the projection on z?: " << ((p_x&p_y&p_z).exists(varset)==p_z) << endl;
  cout << "... while forall(x,y) is ZERO()?: " << ((p_x&p_y&p_z).forall(varset)==SOBDDFunction::ZERO()) << endl;

  // function composition is also supported
  cout << "... function composition is supported? :" << (p_x.compose("x",p_y*p_z)==p_y*p_z) << endl;

  // rename can be applied on functions
  // the effect of g=f.rename(prefix,str) is that every variable starting
  // with the specified prefix will get the new prefix str
  cout << "... renaming the projection on z to z' leads to the projection on z'?: " << (p_z.rename("z","z'")==SOBDDFunction::projection("z'")) << endl;

  // rename can also be done with sets of variables
  VarSet oldVars, newVars;
  oldVars << "x" << "z";
  newVars << "x'" << "z'";
  cout << "... renaming x=>x' and z=>z' in AND(x,AND(y,z)) leads to AND(x',AND(y,z'))?: " << ((p_x&p_y&p_z).rename(oldVars,newVars)==(SOBDDFunction::projection("x'")&p_y&SOBDDFunction::projection("z'"))) << endl;

  // also restrict and constrain are available in the package
  // which lead to the same result in this special case here
  cout << "... constrain and restrict are equal for projection functions?: " << (p_y.restrict(p_z)==p_y.constrain(p_z)) << endl;

  // assignments can easily be used to build multiple cofactors at once
  // another example for assignments can be found in the n-queens problem example
  SOBDDFunction assign=!p_x+(p_x*p_y);
  Assignment assignment1;
  assignment1["x"]=1;
  assignment1["y"]=false;

  cout << "... assignments to more than one variable can be stored and used?: " << (assign.cofactor(assignment1)==SOBDDFunction::ZERO()) << endl;

  Assignment assignment2=assign.satisfyOne();
  cout << "... while assignments from satisfyOne() are always true (if f!=ZERO())?: " << (assign.cofactor(assignment2)==SOBDDFunction::ONE()) << endl;

  cout << "... assignments can also be casted to functions?: " << (SOBDDFunction(assignment1)==(p_x&!p_y)) << endl;

  cout << "... satisfiable assignments can be counted: " << assign.countSatisfiableAssignments() << endl;

  return EXIT_SUCCESS;
}
